"""Client to connect to a database server."""

import socket
from contextlib import contextmanager
from typing import Dict

import numpy.typing as npt

from luxdb.commands import (AddItemsCommand, CountCommand, CreateIndexCommand, DeleteIndexCommand, DeleteItemCommand,
                            GetEFCommand, GetEFConstructionCommand, IndexExistsCommand, InfoCommand, InitIndexCommand,
                            MaxElementsCommand, QueryIndexCommand, ResizeIndexCommand, Result, SetEFCommand)
from luxdb.connection import (receive_result_sync, send_close_sync, send_command_sync)


class SyncClient:
    """Client to connect to a database. This is the synchronous version of the client."""
    def __init__(self, host, port):
        self.host = host
        self.port = port

        self.socket = None

    def connect(self):
        """Connect to the server"""
        self.socket = socket.create_connection((self.host, self.port))

    def send_command(self, command) -> Result:
        """Send the command to the server and return the result the server sends back."""
        send_command_sync(self.socket, command)
        result = receive_result_sync(self.socket)

        return result.get_value()

    def index_exists(self, name: str) -> bool:
        """Check if the index already exists."""
        command = IndexExistsCommand(name=name)
        return self.send_command(command)

    def create_index(self, name: str, space: str, dim: int) -> bool:
        """Create a new index with the given space (l2, ip, cosine) and dimension.

        More information about the parameters is available here:
        https://github.com/nmslib/hnswlib/blob/master/ALGO_PARAMS.md
        """
        command = CreateIndexCommand(name=name, space=space, dim=dim)
        return self.send_command(command)

    def init_index(self, name: str, max_elements: int, ef_construction: int = 200, M: int = 16) -> None:
        """Initialize the index with the max_elements, ef_construction and M.

        More information about the parameters is available here:
        https://github.com/nmslib/hnswlib/blob/master/ALGO_PARAMS.md
        """
        command = InitIndexCommand(name=name, max_elements=max_elements, ef_construction=ef_construction, M=M)
        return self.send_command(command)

    def delete_index(self, name: str) -> None:
        """Delete the index with the given name."""
        command = DeleteIndexCommand(name=name)
        return self.send_command(command)

    def add_items(self, name: str, data: npt.ArrayLike, ids: npt.ArrayLike) -> None:
        """Add the given items to the index."""
        command = AddItemsCommand(name=name, data=data, ids=ids)
        return self.send_command(command)

    def set_ef(self, name: str, new_ef: int):
        """Set the ef to a new value."""
        command = SetEFCommand(name=name, new_ef=new_ef)
        return self.send_command(command)

    def query_index(self, name: str, vector: npt.ArrayLike, k: int) -> npt.ArrayLike:
        """Find the k nearest neighbors of every element in vector.
        Returns a tuple with the labels of the found neighbors and the distances.
        """
        command = QueryIndexCommand(name=name, vector=vector, k=k)
        return self.send_command(command)

    def get_ef(self, name: str) -> int:
        """Get the ef value."""
        command = GetEFCommand(name=name)
        return self.send_command(command)

    def get_ef_construction(self, name: str) -> int:
        """Get the ef construction value"""
        command = GetEFConstructionCommand(name=name)
        return self.send_command(command)

    def delete_item(self, name: str, label: int) -> None:
        """Mark an item as deleted, this will exclude it from search results."""
        command = DeleteItemCommand(name=name, label=label)
        return self.send_command(command)

    def resize_index(self, name: str, new_size: int) -> None:
        """Resize the index to fit more ore less items"""
        command = ResizeIndexCommand(name=name, new_size=new_size)
        return self.send_command(command)

    def count(self, name: str) -> int:
        """Return the current amount of items in the index."""
        command = CountCommand(name=name)
        return self.send_command(command)

    def max_elements(self, name: str) -> int:
        """Return the maximal amount of items"""
        command = MaxElementsCommand(name=name)
        return self.send_command(command)

    def info(self, name: str) -> Dict:
        """Get collection of information about the index.

        Returns a dict with space, dim, M, ef_construction, ef, max_elements, element_count
        """
        command = InfoCommand(name=name)
        return self.send_command(command)

    def quit(self) -> None:
        """Quit the connection and inform the server about it."""
        send_close_sync(self.socket)


@contextmanager
def connect(host, port) -> SyncClient:
    """Provides a context manager for a client connection to host and port"""
    client = SyncClient(host, port)
    client.connect()
    try:
        yield client
    finally:
        client.quit()



"""This module provides command objects that are sent from the client to the server."""
import asyncio
import concurrent.futures
import logging
from enum import Enum, auto
from functools import partial

from luxdb.exceptions import KNNBaseException, NotACommandException
from luxdb.knn_store import KNNStore

LOG = logging.getLogger('commands')


class CommandState(Enum):
    """States that a command can be in."""
    CREATED = auto()
    SENT = auto()
    RECEIVED = auto()
    EXECUTED = auto()
    FAILED = auto()
    SUCCEEDED = auto()


class Result:
    """Wrapper around a result, so that we can also transport state and implement other stuff later on."""
    def __init__(self, data, state):
        self.data = data
        self.state = state

    def raise_error(self):
        """Raise an exception if the result is failed."""
        if self.state == CommandState.FAILED:
            raise self.data

    def get_value(self) -> any:
        """Checks if the result is successfull or not and returns the value if so."""
        self.raise_error()
        return self.data


class Command:
    """Base class for executing remote commands."""
    state = CommandState.CREATED
    result = None

    is_cpu_heavy = False

    def __init__(self, **kwargs):
        self.command_args = kwargs

    async def execute(self, store) -> Result:
        """Executes the command and stores the result or error."""
        self.state = CommandState.EXECUTED
        try:
            if self.is_cpu_heavy:
                LOG.debug('Executing command %r in thread pool.', self)
                loop = asyncio.get_running_loop()
                with concurrent.futures.ThreadPoolExecutor() as pool:
                    self.result = await loop.run_in_executor(pool,
                                                             partial(self.execute_command, store, **self.command_args))
            else:
                LOG.debug('Executing command %r in event loop.', self)
                self.result = self.execute_command(store, **self.command_args)
            self.state = CommandState.SUCCEEDED
        except KNNBaseException as e:
            self.result = e
            self.state = CommandState.FAILED
        except Exception as e:  # pylint: disable=W0703
            logging.getLogger().exception('Unhandled exception %r', e)
            self.result = e
            self.state = CommandState.FAILED

        return Result(self.result, self.state)

    def execute_command(self, store: KNNStore, **kwargs):
        """Commands need to implement this to return the command."""
        raise NotImplementedError('Commands need to provide this method')


def execute_command(command: Command, store: KNNStore):
    """Checks if the command is an instance of command and if so execute it."""
    if isinstance(command, Command):
        return command.execute(store)
    raise NotACommandException('Command is not instance of command!')


class CreateIndexCommand(Command):
    """Create a new index on the receiving storage

    For the parameters refer to the corresponding method in KNNStore.
    """
    def execute_command(self, store: KNNStore, **kwargs):
        return store.create_index(**kwargs)


class InitIndexCommand(Command):
    """Init the index."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.init_index(**kwargs)


class IndexExistsCommand(Command):
    """Check if the index exists."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.index_exists(**kwargs)


class DeleteIndexCommand(Command):
    """Delete an index."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.delete_index(**kwargs)


class AddItemsCommand(Command):
    """Add items to the specified index."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.add_items(**kwargs)


class SetEFCommand(Command):
    """Set the ef."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.set_ef(**kwargs)


class GetEFCommand(Command):
    """Get the ef."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.get_ef(**kwargs)


class GetEFConstructionCommand(Command):
    """get construction ef"""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.get_ef_construction(**kwargs)


class QueryIndexCommand(Command):
    """Query the index."""
    is_cpu_heavy = True

    def execute_command(self, store: KNNStore, **kwargs):
        return store.query_index(**kwargs)


class DeleteItemCommand(Command):
    """Delete an item from the index."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.delete_item(**kwargs)


class ResizeIndexCommand(Command):
    """Resize the index to a new size"""
    is_cpu_heavy = True

    def execute_command(self, store: KNNStore, **kwargs):
        return store.resize_index(**kwargs)


class MaxElementsCommand(Command):
    """Get the current limit on the index."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.max_elements(**kwargs)


class CountCommand(Command):
    """Get the current amount of items."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.count(**kwargs)


class InfoCommand(Command):
    """Get information about the index."""
    def execute_command(self, store: KNNStore, **kwargs):
        return store.info(**kwargs)



"""Helper functions to send and receive objects over socket streams."""
import asyncio
import logging
import pickle  # nosec
import socket
from typing import Tuple

from luxdb.commands import Command, Result

LOG = logging.getLogger('connections')


def pack_obj(obj) -> Tuple[int, bytes]:
    """Packs an object for sending and returns the size and the data"""
    data = pickle.dumps(obj)
    size = len(data)
    return size, data


async def send_obj(writer: asyncio.StreamWriter, obj: any) -> None:
    """Send object over the writer. first the size and after that the data."""
    size, data = pack_obj(obj)

    LOG.debug('Sending %d bytes of data.', size)

    writer.write(size.to_bytes(8, 'big'))
    writer.write(data)

    await writer.drain()


def send_obj_sync(writer: socket.socket, obj: any) -> None:
    """Sends an object to another socket synchronously."""
    size, data = pack_obj(obj)

    writer.sendall(size.to_bytes(8, 'big'))
    writer.sendall(data)


async def receive_obj(reader: asyncio.StreamReader) -> any:
    """Receive an object over the reader."""
    size = int.from_bytes(await reader.readexactly(8), 'big')

    if size == 0:
        LOG.info('Command with size 0 received, return None')
        return None

    LOG.debug('Expecting %d bytes of data', size)
    data = await reader.readexactly(size)
    LOG.debug('Received %d bytes of data', size)

    obj = pickle.loads(data)  # nosec
    #obj = json.loads(data)

    return obj


def receive_obj_sync(reader: socket.socket) -> any:
    """Receive an object over the socket synchronously."""
    size = int.from_bytes(reader.recv(8, socket.MSG_WAITALL), 'big')

    if size == 0:
        LOG.info('Command with size 0 received, return None')
        return None

    data = reader.recv(size, socket.MSG_WAITALL)
    obj = pickle.loads(data)  # nosec

    return obj


async def receive_command(reader: asyncio.StreamReader) -> Command:
    """Waits for the sender to send a command object."""
    cmd = await receive_obj(reader)

    if isinstance(cmd, Command):
        return cmd

    if cmd is None:
        return None

    raise Exception(f'Received unknown object: {cmd}')


async def receive_result(reader: asyncio.StreamReader) -> Result:
    """Receive a result over the reader."""
    result = await receive_obj(reader)

    if isinstance(result, Result):
        return result

    raise Exception(f'Received unknown object: {result}')


def receive_result_sync(reader: socket.socket) -> Result:
    """Receive a result over the reader synchronously."""
    result = receive_obj_sync(reader)

    if isinstance(result, Result):
        return result

    raise Exception(f'Received unknown object: {result}')


async def send_command(writer: asyncio.StreamWriter, cmd: Command) -> None:
    """Send the command over the writer."""
    await send_obj(writer, cmd)


def send_command_sync(writer: socket.socket, cmd: Command) -> None:
    """Send the command over the writer synchronously."""
    send_obj_sync(writer, cmd)


async def send_result(writer: asyncio.StreamWriter, result: Result) -> None:
    """Send the result over the writer."""
    await send_obj(writer, result)


def send_result_sync(writer: socket.socket, result: Result) -> None:
    'Send the result over the writer synchronously.'
    send_obj_sync(writer, result)


async def send_close(writer: asyncio.StreamWriter) -> None:
    """Sends an empty object to signal closing of the connection and then close the writer."""
    writer.write((0).to_bytes(8, 'big'))
    await writer.drain()
    writer.close()

    await writer.wait_closed()


def send_close_sync(writer: socket.socket) -> None:
    """Sends an empty object to signal closing of the connection and then close the writer synchronously."""
    writer.sendall((0).to_bytes(8, 'big'))
    writer.close()



"""Client to connect to a database server."""

import asyncio
from contextlib import asynccontextmanager
from typing import Dict

import numpy.typing as npt

from luxdb.commands import (AddItemsCommand, CountCommand, CreateIndexCommand, DeleteIndexCommand, DeleteItemCommand,
                            GetEFCommand, GetEFConstructionCommand, IndexExistsCommand, InfoCommand, InitIndexCommand,
                            MaxElementsCommand, QueryIndexCommand, ResizeIndexCommand, Result, SetEFCommand)
from luxdb.connection import receive_result, send_close, send_command


class Client:
    """Client to connect to a database."""
    def __init__(self, host, port):
        self.host = host
        self.port = port

        self.reader = None
        self.writer = None

    async def connect(self):
        """Connect to the server"""
        self.reader, self.writer = await asyncio.open_connection(self.host, self.port)

    async def send_command(self, command) -> Result:
        """Send the command to the server and return the result the server sends back."""
        await send_command(self.writer, command)
        result = await receive_result(self.reader)

        return result.get_value()

    async def index_exists(self, name: str) -> bool:
        """Check if the index already exists."""
        command = IndexExistsCommand(name=name)
        return await self.send_command(command)

    async def create_index(self, name: str, space: str, dim: int) -> bool:
        """Create a new index with the given space (l2, ip, cosine) and dimension.

        More information about the parameters is available here:
        https://github.com/nmslib/hnswlib/blob/master/ALGO_PARAMS.md
        """
        command = CreateIndexCommand(name=name, space=space, dim=dim)
        return await self.send_command(command)

    async def init_index(self, name: str, max_elements: int, ef_construction: int = 200, M: int = 16) -> None:
        """Initialize the index with the max_elements, ef_construction and M.

        More information about the parameters is available here:
        https://github.com/nmslib/hnswlib/blob/master/ALGO_PARAMS.md
        """
        command = InitIndexCommand(name=name, max_elements=max_elements, ef_construction=ef_construction, M=M)
        return await self.send_command(command)

    async def delete_index(self, name: str) -> None:
        """Delete the index with the given name."""
        command = DeleteIndexCommand(name=name)
        return await self.send_command(command)

    async def add_items(self, name: str, data: npt.ArrayLike, ids: npt.ArrayLike) -> None:
        """Add the given items to the index."""
        command = AddItemsCommand(name=name, data=data, ids=ids)
        return await self.send_command(command)

    async def set_ef(self, name: str, new_ef: int):
        """Set the ef to a new value."""
        command = SetEFCommand(name=name, new_ef=new_ef)
        return await self.send_command(command)

    async def query_index(self, name: str, vector: npt.ArrayLike, k: int) -> npt.ArrayLike:
        """Find the k nearest neighbors of every element in vector.
        Returns a tuple with the labels of the found neighbors and the distances.
        """
        command = QueryIndexCommand(name=name, vector=vector, k=k)
        return await self.send_command(command)

    async def get_ef(self, name: str) -> int:
        """Get the ef value."""
        command = GetEFCommand(name=name)
        return await self.send_command(command)

    async def get_ef_construction(self, name: str) -> int:
        """Get the ef construction value"""
        command = GetEFConstructionCommand(name=name)
        return await self.send_command(command)

    async def delete_item(self, name: str, label: int) -> None:
        """Mark an item as deleted, this will exclude it from search results."""
        command = DeleteItemCommand(name=name, label=label)
        return await self.send_command(command)

    async def resize_index(self, name: str, new_size: int) -> None:
        """Resize the index to fit more ore less items"""
        command = ResizeIndexCommand(name=name, new_size=new_size)
        return await self.send_command(command)

    async def count(self, name: str) -> int:
        """Return the current amount of items in the index."""
        command = CountCommand(name=name)
        return await self.send_command(command)

    async def max_elements(self, name: str) -> int:
        """Return the maximal amount of items"""
        command = MaxElementsCommand(name=name)
        return await self.send_command(command)

    async def info(self, name: str) -> Dict:
        """Get collection of information about the index.

        Returns a dict with space, dim, M, ef_construction, ef, max_elements, element_count
        """
        command = InfoCommand(name=name)
        return await self.send_command(command)

    async def quit(self) -> None:
        """Quit the connection and inform the server about it."""
        await send_close(self.writer)


@asynccontextmanager
async def connect(host, port) -> Client:
    """Provides a context manager for a client connection to host and port"""
    client = Client(host, port)
    await client.connect()
    try:
        yield client
    finally:
        await client.quit()



"""This module implements a storage for multi-dimensional vectors and fast nearest neighbor search.

It wraps around https://github.com/nmslib/hnswlib and provides thread safe access to the operations.

The Store can also be saved and loaded onto the disk.
"""
import logging
from contextlib import contextmanager
from typing import Dict

import hnswlib
import numpy.typing as npt
import persistent
import transaction
import ZODB
import ZODB.FileStorage
from BTrees.OOBTree import \
    OOBTree  # pylint: disable=import-error,no-name-in-module

from luxdb.exceptions import (IndexAlreadyExistsException, IndexDoesNotExistException, UnknownSpaceException)
from luxdb.index import Index

LOG = logging.getLogger('store')


class KNNStore(persistent.Persistent):
    """
    Store multiple indexes with different attributes.
    """

    ALLOWED_SPACES = ['l2', 'ip', 'cosine']

    def __init__(self, path: str = None, storage=None):
        """Create the database under the specified path or with the given storage.

        If `None` is given the database will be created in memory only.
        """
        self.transaction = transaction.TransactionManager()
        if path is not None:
            LOG.debug('Connecting to storage at %s', path)
            self.storage = ZODB.FileStorage.FileStorage(path)
        else:
            self.storage = storage

        self.db = ZODB.DB(self.storage)

        if self.storage is None:
            LOG.debug('Creating database in memory.')
            self.load_database()
        else:
            self.root = None

        self.path = path

    def get_index(self, name: str) -> hnswlib.Index:
        """Access the index with the given name."""
        try:
            return self.root['indexes'][name]
        except KeyError as e:
            raise IndexDoesNotExistException(name) from e

    def load_database(self):
        """Initialize the database from the filesystem"""
        LOG.debug('Loading database.')
        connection = self.db.open(self.transaction)
        self.root = connection.root()

        if not 'indexes' in self.root:
            LOG.info('DB does not exist, creating empty database.')
            self.root['indexes'] = OOBTree()
            self.transaction.commit()

        LOG.debug('Loaded database.')

    def close(self):
        """Closes the connection to the database so the locks are released."""
        self.db.close()

    def index_exists(self, name: str) -> bool:
        """Returns true if the index already exists."""
        LOG.debug('Index %s exists: %s', name, name in self.root['indexes'])
        return name in self.root['indexes']

    def create_index(self, name: str, space: str, dim: int) -> bool:
        """Create a new index with the given name and parameters."""
        with self.transaction:
            if space not in KNNStore.ALLOWED_SPACES:
                raise UnknownSpaceException(space)

            LOG.debug('Creating new index: %s, space: %s, dim: %d', name, space, dim)
            if name in self.root['indexes']:
                LOG.error('Index with name %s already exists', name)
                raise IndexAlreadyExistsException(name)

            index = hnswlib.Index(space=space, dim=dim)

            self.root['indexes'][name] = Index(index)

            LOG.debug('New index created: %s', name)
            return True

    def init_index(self, name: str, max_elements: int, ef_construction: int = 200, M: int = 16) -> None:
        """Init the index with the given parameters.

        More information about the parameters is available here:
        https://github.com/nmslib/hnswlib/blob/master/ALGO_PARAMS.md
        """
        LOG.debug('Initializing index %s with max_elements: %d, ef_construction: %d, M: %d', name, max_elements,
                  ef_construction, M)

        with self.transaction:
            index = self.get_index(name)
            index.init_index(
                max_elements=max_elements,
                ef_construction=ef_construction,
                M=M,
            )

        LOG.info('Initialized index %s', name)

    def delete_index(self, name: str) -> None:
        """Deletes the index from the store."""
        LOG.debug('Deleting index %s', name)
        with self.transaction:
            del self.root['indexes'][name]
            LOG.info('Deleted index %s', name)

    def add_items(self, name: str, data: npt.ArrayLike, ids: npt.ArrayLike):
        """Add the items with the ids to the index."""
        LOG.debug('Adding items to %s', name)
        with self.transaction:
            index = self.get_index(name)
            index.add_items(data, ids)

        LOG.debug('Added items to %s', name)

    def set_ef(self, name: str, new_ef: int):
        """Sets the ef value on the index."""
        with self.transaction:
            index = self.get_index(name)

            index.set_ef(new_ef)

    def get_ef(self, name: str) -> int:
        """Return the current value of ef"""
        return self.get_index(name).ef

    def get_ef_construction(self, name: str) -> int:
        """Return the value of construction_ef"""
        return self.get_index(name).ef_construction

    def query_index(self, name: str, vector: npt.ArrayLike, k: int = 1) -> npt.NDArray:
        """Return the k nearest vectors to the given vector"""
        index = self.get_index(name)
        return index.knn_query(vector, k)

    def delete_item(self, name: str, label: int) -> None:
        """This does just remove the item from search results, it's still in the index."""
        with self.transaction:
            index = self.get_index(name)

            index.mark_deleted(label)

    def resize_index(self, name: str, new_size: int) -> None:
        """Resize the index to accommodate more or less items."""
        with self.transaction:
            index = self.get_index(name)
            index.resize_index(new_size)

    def max_elements(self, name: str) -> int:
        """Return the maximum number of elements that can be stored in the index."""
        return self.get_index(name).get_max_elements()

    def count(self, name: str) -> int:
        """Return the current amount of elements in the index."""
        return self.get_index(name).get_current_count()

    def info(self, name: str) -> Dict:
        """Get information about the index."""
        index = self.get_index(name)
        return {
            'space': index.space,
            'dim': index.dim,
            'M': index.M,
            'ef_construction': index.ef_construction,
            'max_elements': index.get_max_elements(),
            'element_count': index.get_current_count(),
            'ef': index.ef,
        }


@contextmanager
def open_store(path):
    """Open a store, load it if possible and save it as soon as the context is over."""
    store = KNNStore(path)
    store.load_database()

    try:
        yield store
    finally:
        store.close()



"""Implementation of persistence."""
import logging
from contextlib import contextmanager
from threading import Condition, Lock, current_thread
from typing import List

import hnswlib
import persistent

LOG = logging.getLogger('store')


class ExclusiveLock:
    """
    This implements a lock specific for the library we are using.

    If one thread is in A, no threads may be in B or in C, but other threads may enter A.
    As soon as all threads have left A, other threads can enter B or C. As soon as one thread enters
    B, no other threads can enter A or C.

    """
    def __init__(self, states: List[str]):
        self.state_lock = Lock()
        self.condition = Condition(self.state_lock)

        self.locks = {state: Lock() for state in states}
        self.counters = {state: 0 for state in states}

    @contextmanager
    def __call__(self, tag):
        self.acquire(tag)
        try:
            yield self
        finally:
            self.release(tag)

    def acquire(self, requested_state: str):
        """
        If the path is free this will make sure that all other locks are locked.
        Otherwise wait until the path becomes free.
        """
        LOG.debug('Acquiring state lock for %s (Thread: %s)', requested_state, current_thread().ident)
        with self.state_lock:
            LOG.debug('Got state_lock for locking for %s (Thread: %s)', requested_state, current_thread().ident)

            self.condition.wait_for(lambda: self.locks[requested_state].acquire(blocking=False))

            LOG.debug('Got state_lock for %s (Thread: %s)', requested_state, current_thread().ident)

            self.counters[requested_state] += 1
            LOG.debug('State counter for %s is %d (Thread: %s)', requested_state, self.counters[requested_state],
                      current_thread().ident)

            for key, lock in self.locks.items():
                LOG.debug('Locking other locks for state %s (Thread: %s)', requested_state, current_thread().ident)
                if key != requested_state and not lock.locked():
                    LOG.debug('Acquiring lock for %s (Thread: %s)', key, current_thread().ident)
                    lock.acquire()

            LOG.debug('State locks are setup, releasing lock for %s (Thread: %s)', requested_state,
                      current_thread().ident)
            self.locks[requested_state].release()
            LOG.debug('Lock acquired for %s (Thread: %s)', requested_state, current_thread().ident)

    def release(self, requested_state: str):
        """Releases all other locks, as soon as all threads from the same state have released the lock."""
        LOG.debug('Releasing lock for state %s (Thread: %s)', requested_state, current_thread().ident)

        with self.state_lock:
            LOG.debug('Got state_lock for releasing of %s (Thread: %s)', requested_state, current_thread().ident)
            self.condition.wait_for(lambda: self.locks[requested_state].acquire(blocking=False))

            self.counters[requested_state] -= 1
            LOG.debug('State counter for %s is %d (Thread: %s)', requested_state, self.counters[requested_state],
                      current_thread().ident)

            if self.counters[requested_state] == 0:
                LOG.debug('Releasing locks for the other states. (Thread: %s)', current_thread().ident)

                for key, lock in self.locks.items():
                    LOG.debug('Trying to release lock for %s (Thread: %s)', key, current_thread().ident)
                    if key != requested_state and lock.locked():
                        LOG.debug('Unlocking lock %s (Thread: %s)', requested_state, current_thread().ident)
                        lock.release()
                        LOG.debug('Lock released for %s (Thread: %s)', key, current_thread().ident)
                self.condition.notify_all()

            self.locks[requested_state].release()
            LOG.debug('Released lock for state %s', requested_state)


class Index(persistent.Persistent):
    """Wrapper around hnswlib.Index to record accesses and store changes."""

    EXCLUSIVE_STATES = ['add', 'resize', 'query']

    def __init__(self, index: hnswlib.Index):
        self.index = index
        self._v_lock = ExclusiveLock(self.EXCLUSIVE_STATES)

    @property
    def lock(self):
        """Return the lock that is prefixed with _v_ so it's not persisted."""
        if not hasattr(self, '_v_lock'):  # Not initialized if read from db
            self._v_lock = ExclusiveLock(self.EXCLUSIVE_STATES)
        return self._v_lock

    def _mark_change(self):
        """Mark the index as changed since the latest commit."""
        self._p_changed = True

    def init_index(self, max_elements: int, ef_construction: int = 200, M: int = 16):
        """Inits the index and marks the index as changed."""
        self.index.init_index(
            max_elements=max_elements,
            ef_construction=ef_construction,
            M=M,
        )
        self._mark_change()

    def add_items(self, data, ids):
        """Add items and mark index as changed."""
        with self.lock('add'):
            self.index.add_items(data, ids)
            self._mark_change()

    def set_ef(self, new_ef):
        """Set the the ef."""
        self.index.set_ef(new_ef)
        self._mark_change()

    @property
    def ef(self):
        """Access the ef of the index."""
        return self.index.ef

    @property
    def ef_construction(self):
        """Access the ef_construction of the index"""
        return self.index.ef_construction

    @property
    def space(self):
        """Access the space of the index."""
        return self.index.space

    @property
    def dim(self):
        """Access the dim of the index."""
        return self.index.dim

    @property
    def M(self):
        """Access the M of the index."""
        return self.index.M

    def knn_query(self, vector, k):
        """Query the index."""
        with self.lock('query'):
            return self.index.knn_query(vector, k)

    def resize_index(self, new_size):
        """Resize the index and mark the index."""
        with self.lock('resize'):
            self.index.resize_index(new_size)
            self._mark_change()

    def mark_deleted(self, label):
        """Mark an item as deleted."""
        self.index.mark_deleted(label)
        self._mark_change()

    def get_max_elements(self):
        """Get the maximum number of elements for the index."""
        return self.index.get_max_elements()

    def get_current_count(self):
        """Get the current number of elements in the index."""
        return self.index.get_current_count()



"""Exceptions in the LuxDB"""


class KNNBaseException(Exception):
    """Base class for exceptions in this module"""


class IndexAlreadyExistsException(KNNBaseException):
    """Thrown if an index already exists after creation was requested."""
    def __init__(self, name: str):
        """Create a new Exception for the given name."""
        self.name = name
        super().__init__(name)


class IndexDoesNotExistException(KNNBaseException):
    """The requested index does not exist in the database."""
    def __init__(self, name: str):
        self.name = name
        super().__init__(name)


class UnknownSpaceException(KNNBaseException):
    """Space is not known"""
    def __init__(self, space):
        self.space = space
        super().__init__(space)


class NotACommandException(KNNBaseException):
    """Tried to execute something that is not a command."""
    def __init__(self, obj):
        self.obj = obj
        super().__init__(obj)



"""This module provides a simple tcp server to connect to the database."""
import argparse
import asyncio
import logging
import socket

from luxdb.connection import receive_command, send_result
from luxdb.knn_store import KNNStore, open_store

LOG = logging.getLogger('server')


class Server:
    """TCP Server that allows clients to connect to the database.

    Receives commands, executes them on the database and sends the result back.

    """
    def __init__(self, host: str, port: int, store: KNNStore):
        """Define the host and port where the server will listen, and the store that should be used."""
        self.host = host
        self.port = port
        self.store = store
        self.server = None

    async def connection(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """Handles a single incoming connection asynchronously"""

        command = await receive_command(reader)

        while command is not None:
            result = await command.execute(self.store)

            await send_result(writer, result)

            command = await receive_command(reader)

        writer.close()
        await writer.wait_closed()

    async def create_server(self):
        """Creates the server socket but doesn't start listening yet."""
        self.server = await asyncio.start_server(self.connection, self.host, self.port)

        for server_socket in self.server.sockets:
            host, port = socket.getnameinfo(server_socket.getsockname(), socket.NI_NUMERICSERV | socket.NI_NUMERICHOST)
            LOG.info('Serving on %s:%s', host, port)
            self.host = host
            self.port = int(port)

    async def start(self, callback=None):
        """Start the server and process incoming connections.

        The callback is called shortly before the server starts listening, you can use it to notify
        clients that the server is listening, so they can connect without an error.
        """
        if self.server is None:
            await self.create_server()

        async with self.server:
            if callback is not None:
                asyncio.get_running_loop().call_soon(callback)
            await self.server.serve_forever()


async def serve(args: dict):
    """Main method for the server thread."""

    with open_store(args.path) as store:
        server = Server(host=args.host, port=args.port, store=store)
        try:
            await server.start()
        except KeyboardInterrupt:
            print('Shutting down server..')


def main():
    """Main method for the server."""
    try:
        parser = argparse.ArgumentParser(description='Multidimensional vector database (server).')

        parser.add_argument('--host', type=str, default='127.0.0.1', help='Host where the server should listen.')
        parser.add_argument('--port', type=int, default=None, help='Port to listen on.')
        parser.add_argument(
            '-log',
            '--loglevel',
            default='warning',
            help='Provide logging level. Example --loglevel debug, default=warning',
        )
        parser.add_argument('path', type=str, help='Path where the database is stored or should be stored.')
        args = parser.parse_args()
        logging.basicConfig(level=args.loglevel.upper())

        asyncio.run(serve(args), debug=args.loglevel.upper() == 'DEBUG')
    except KeyboardInterrupt:
        print('Shutting down server...')
